`include "opcode.vh"

module control (
  input [31:0] instr_fd,
  input [31:0] instr_ex,
  input [31:0] instr_wb,
  input BrEq,
  input BrLT,

  output wire ALUSel,
  output wire ASel,
  output wire BSel,
  output reg [1:0] RS1Sel,
  output reg [1:0] RS2Sel,
  output wire RS1SelEx,
  output wire RS2SelEx,
  output wire RegWEn,
  output wire MemRW,
  output reg [1:0] WBSel,
  output reg BrUn,
  output reg InsertNop,
  output reg [1:0] PCSel
);

  wire [6:0] opcode = instr_fd[6:0];
  wire [2:0] funct3 = instr_fd[14:12];
  wire [6:0] funct7 = instr_fd[31:25];

  wire [6:0] opcode_ex = instr_ex[6:0];
  wire [6:0] funct7_ex = instr_ex[31:25];

  wire [6:0] opcode_wb = instr_wb[6:0];

  wire instr_fd_has_rs1;
  wire instr_fd_has_rs2;
  wire instr_ex_has_rs1;
  wire instr_ex_has_rs2;
  wire instr_wb_has_rd;
  wire instr_ex_has_rd;

  reg branch_taken = 1'b0;
  
  reg bit conflict = 1'b0;

  //assign ALUSel = (instr_ex[31:25] == 7'h20) ? 1'b1 : 1'b0; // Prior to AUIPC
  assign ALUSel = ((opcode_ex == `OPC_ARI_RTYPE ||
                   opcode_ex == `OPC_ARI_ITYPE) &&
                   funct7_ex == 7'h20)          ? 1'b1 : 1'b0;
  assign ASel   = (opcode_ex == `OPC_BRANCH     ||
                   opcode_ex == `OPC_AUIPC      ||
                   opcode_ex == `OPC_JAL)       ? 1'b1 : 1'b0;
  assign BSel   = (opcode_ex == `OPC_ARI_RTYPE) ? 1'b0 : 1'b1;
  assign RegWEn = (opcode_wb != `OPC_STORE      &&
                   opcode_wb != `OPC_BRANCH     &&
                   instr_wb[11:7] != 0)         ? 1'b1 : 1'b0;
  assign MemRW  = (opcode_ex == `OPC_STORE)     ? 1'b1 : 1'b0;
  /*assign WBSel = (opcode_wb == `OPC_ARI_RTYPE 
                  || opcode_wb == `OPC_ARI_ITYPE
                  || opcode_wb == `OPC_LUI
                  || opcode_wb == `OPC_AUIPC) ? 2'b01 : (opcode_wb == `OPC_JAL) ? 2'b10 : 2'b00;*/

  assign instr_fd_has_rs1 = (opcode    != `OPC_JAL    &&
                             opcode    != `OPC_LUI    &&
                             opcode    != `OPC_AUIPC) ? 1'b1 : 1'b0;

  assign instr_ex_has_rs1 = (opcode_ex != `OPC_JAL    &&
                             opcode_ex != `OPC_LUI    &&
                             opcode_ex != `OPC_AUIPC) ? 1'b1 : 1'b0;

  assign instr_fd_has_rs2 = (opcode == `OPC_ARI_RTYPE || opcode == `OPC_STORE || opcode == `OPC_BRANCH) ? 1'b1 : 1'b0;
  assign instr_ex_has_rs2 = (opcode_ex == `OPC_ARI_RTYPE || opcode_ex == `OPC_STORE || opcode_ex == `OPC_BRANCH) ? 1'b1 : 1'b0;

  assign instr_ex_has_rd  = (opcode_ex != `OPC_STORE && opcode_ex != `OPC_BRANCH && instr_ex[11:7] != 0) ? 1'b1 : 1'b0;
  assign instr_wb_has_rd  = (opcode_wb != `OPC_STORE && opcode_wb != `OPC_BRANCH && instr_wb[11:7] != 0) ? 1'b1 : 1'b0;

  always @* begin
    if (instr_fd_has_rs1) begin
      if ((instr_ex[11:7] == instr_fd[19:15]) && instr_ex_has_rd) RS1Sel = 2'b01;
      else if      ((instr_wb[11:7] == instr_fd[19:15]) && instr_wb_has_rd) RS1Sel = 2'b10;
      else RS1Sel = 2'b00;
    end

    if (instr_fd_has_rs2) begin
      if ((instr_ex[11:7] == instr_fd[24:20]) && instr_ex_has_rd) RS2Sel = 2'b01;
      else if      ((instr_wb[11:7] == instr_fd[24:20]) && instr_wb_has_rd) RS2Sel = 2'b10;
      else RS2Sel = 2'b00;
    end
  end

  assign RS1SelEx = ((instr_wb[11:7] == instr_ex[19:15]) && instr_ex_has_rs1 && instr_wb_has_rd) ?
                      1'b1 : 1'b0;

  assign RS2SelEx = ((instr_wb[11:7] == instr_ex[24:20]) && instr_ex_has_rs2 && instr_wb_has_rd) ?
                      1'b1 : 1'b0;

  assign InsertNop = ((RS1Sel == 2'b01 || RS2Sel ==2'b01) && opcode ==`OPC_BRANCH && opcode_ex == `OPC_LOAD) ? 1'b1 : 1'b0;

  always @* begin
    case (opcode_wb)
      `OPC_ARI_RTYPE: WBSel = 2'b01;
      `OPC_ARI_ITYPE: WBSel = 2'b01;
      `OPC_LUI:       WBSel = 2'b01;
      `OPC_AUIPC:     WBSel = 2'b01;
      `OPC_JAL:       WBSel = 2'b10;
      `OPC_JALR:      WBSel = 2'b10;
      default:        WBSel = 2'b00;
    endcase
  end

  assign BrUn = (funct3 == 3'b110 || funct3 == 3'b111) ? 1'b1 : 1'b0;
  //assign PCSel = branch_taken;

  always @* begin
    if (InsertNop) PCSel = 2'b11;
    else if (opcode == `OPC_BRANCH) begin
      case (funct3)
        `FNC_BEQ:  PCSel = (BrEq)  ? 2'b01 : 2'b00;
        `FNC_BNE:  PCSel = (!BrEq) ? 2'b01 : 2'b00;
        `FNC_BLT:  PCSel = (BrLT)  ? 2'b01 : 2'b00;
        `FNC_BGE:  PCSel = (!BrLT) ? 2'b01 : 2'b00;
        `FNC_BLTU: PCSel = (BrLT)  ? 2'b01 : 2'b00;
        `FNC_BGEU: PCSel = (!BrLT) ? 2'b01 : 2'b00;
        default: PCSel = 2'b00;
      endcase
    end
    else if (opcode == `OPC_JAL)  PCSel = 2'b01;
    else if (opcode == `OPC_JALR) PCSel = 2'b10;
    else PCSel = 2'b00;
  end 

endmodule
