`include "opcode.vh"

module cpu #(
    parameter CPU_CLOCK_FREQ = 50_000_000,
    parameter RESET_PC = 32'h4000_0000,
    parameter BAUD_RATE = 115200
) (
    input clk,
    input rst
);
    // BIOS Memory
    // Synchronous read: read takes one cycle
    // Synchronous write: write takes one cycle
    wire [11:0] bios_addra, bios_addrb;
    wire [31:0] bios_douta, bios_doutb;
    wire bios_ena, bios_enb;
    bios_mem bios_mem (
      .clk(clk),
      .ena(bios_ena),
      .addra(bios_addra),
      .douta(bios_douta),
      .enb(bios_enb),
      .addrb(bios_addrb),
      .doutb(bios_doutb)
    );

    // Data Memory
    // Synchronous read: read takes one cycle
    // Synchronous write: write takes one cycle
    // Write-byte-enable: select which of the four bytes to write
    wire [13:0] dmem_addr;
    wire [31:0] dmem_din, dmem_dout;
    wire [3:0] dmem_we;
    wire dmem_en;
    dmem dmem (
      .clk(clk),
      .en(dmem_en),
      .we(dmem_we),
      .addr(dmem_addr),
      .din(dmem_din),
      .dout(dmem_dout)
    );

    // Instruction Memory
    // Synchronous read: read takes one cycle
    // Synchronous write: write takes one cycle
    // Write-byte-enable: select which of the four bytes to write
    wire [31:0] imem_dina, imem_doutb;
    wire [13:0] imem_addra, imem_addrb;
    wire [3:0] imem_wea;
    wire imem_ena;
    imem imem (
      .clk(clk),
      .ena(imem_ena),
      .wea(imem_wea),
      .addra(imem_addra),
      .dina(imem_dina),
      .addrb(imem_addrb),
      .doutb(imem_doutb)
    );

    // Register file
    // Asynchronous read: read data is available in the same cycle
    // Synchronous write: write takes one cycle
    wire we;
    wire [4:0] ra1, ra2, wa;
    wire [31:0] wd;
    wire [31:0] rd1, rd2;
    reg_file rf (
        .clk(clk),
        .we(we),
        .ra1(ra1), .ra2(ra2), .wa(wa),
        .wd(wd),
        .rd1(rd1), .rd2(rd2)
    );

  reg [31:0] tohost_csr = 0;

  reg [31:0] pc_fd;
  reg [31:0] pc_ex;
  reg [31:0] pc_wb;
  reg [31:0] next_pc;
  reg [31:0] prev_instr = 0;

  reg [31:0] instr_fd;
  reg [31:0] instr_ex;
  reg [31:0] instr_wb;

  reg [31:0] rd1_fd;
  reg [31:0] rd2_fd;
  reg [31:0] rd1_ex;
  reg [31:0] rd2_ex;
  reg [31:0] rd1_wb;
  reg [31:0] rd2_wb;
  wire [31:0] rd1_fwd_ex;
  wire [31:0] rd2_fwd_ex;

  reg [31:0] alu_ex;
  reg [31:0] alu_wb;

  reg [31:0] imm_ex;
  reg [31:0] imm_wb;

    wire [31:0] alu_in_a, alu_in_b, alu_out;
    wire [3:0] funct3;
    wire alu_sel;
    alu alu (
        .in_a(alu_in_a),
        .in_b(alu_in_b),
        .funct3(funct3),
        .alu_sel(alu_sel),
        .out(alu_out)
    );

    wire alu_a_sel;
    wire alu_b_sel;
    wire [1:0] rs1_sel;
    wire [1:0] rs2_sel;
    wire rs1_sel_ex;
    wire rs2_sel_ex;
    wire insert_nop;
    wire mem_rw;
    wire [1:0] wb_sel;
    wire br_unsigned_comp;
    wire branch_eq;
    wire branch_lt;
    wire [1:0] next_pc_sel;
    control control (
        .instr_fd(instr_fd),
        .instr_ex(instr_ex),
        .instr_wb(instr_wb),
        .BrEq(branch_eq),
        .BrLT(branch_lt),
        .ALUSel(alu_sel),
        .ASel(alu_a_sel),
        .BSel(alu_b_sel),
        .RS1Sel(rs1_sel),
        .RS2Sel(rs2_sel),
        .RS1SelEx(rs1_sel_ex),
        .RS2SelEx(rs2_sel_ex),
        .RegWEn(we),
        .MemRW(mem_rw),
        .WBSel(wb_sel),
        .BrUn(br_unsigned_comp),
        .InsertNop(insert_nop),
        .PCSel(next_pc_sel)
    );

  wire [3:0] wr_mask;
  gen_wr_mask mask_generator (
        .instr(instr_ex),
        .addr(alu_out),
        .wr_mask(wr_mask)
  );

  wire [31:0] wr_data;
  gen_wr_data wr_data_generator (
    .wr_data_in(rd2_fwd_ex),
    .wr_mask(wr_mask),
    .wr_data(wr_data)
  );
  
  wire [31:0] load_extended_data;
  load_extender load_extender (
    .data_in(dmem_dout),
    .funct3(instr_wb[14:12]), // TODO: Move to LDSel on Control Unit
    .addr(alu_wb[1:0]),       // TODO: Move to ByteSel on Control Uni
    .data_out(load_extended_data)
  );

  branch_comparator branch_comparator (
    .rs1(rd1_fd),
    .rs2(rd2_fd),
    .unsigned_comp(br_unsigned_comp),
    .branch_eq(branch_eq),
    .branch_lt(branch_lt)
  );
  
	reg [32:0] clk_count;
  reg [32:0] instr_count;


  reg [31:0] immediate;

  wire [6:0] opcode = instr_fd[6:0];

  // Count clock cycles
	/*always @(posedge clk) begin
		clk_count = clk_count + 1;
	end*/

  // Immediate Generator
  always @(*) begin
    case (instr_fd[6:0])
      `OPC_ARI_RTYPE: immediate = 32'h0000_0000;
      `OPC_ARI_ITYPE: immediate = {{21{instr_fd[31]}}, instr_fd[30:20]};
      `OPC_LOAD:      immediate = {{21{instr_fd[31]}}, instr_fd[30:20]};
      `OPC_STORE:     immediate = {{20{instr_fd[31]}}, instr_fd[30:25], instr_fd[11:7]};
      `OPC_BRANCH:    immediate = {{20{instr_fd[31]}}, instr_fd[7]    , instr_fd[30:25],
                                  instr_fd[11:8], 1'b0};
      `OPC_LUI:       immediate = {{instr_fd[31:12]  , 12'b0}};
      `OPC_AUIPC:     immediate = {{instr_fd[31:12]  , 12'b0}};
      `OPC_JAL:       immediate = {{12{instr_fd[31]}}, instr_fd[19:12], instr_fd[20],
                                  instr_fd[30:25]    , instr_fd[24:21], 1'b0};
      `OPC_JALR:      immediate = {{21{instr_fd[31]}}, instr_fd[30:20]};
      `OPC_CSR:       immediate = {{27{1'b0}}        , instr_fd[19:15]};
      default:        immediate = 32'b0;
    endcase
  end

  // Configure BIOS inputs
  assign bios_addra = next_pc[13:2];
  assign bios_addrb = alu_out;
  assign bios_ena   = (pc_fd[30]) ? 1'b1 : 1'b0;

  // Configure Instruction Memory inputs 
  assign imem_addra = alu_out[15:2];
  assign imem_addrb = next_pc[15:2];
  assign imem_dina  = wr_data;

  // Configure Register File inputs 
  assign wa  = instr_wb[11:7];
  assign ra1 = instr_fd[19:15];
  assign ra2 = instr_fd[24:20];

  // Configure Muxes
  assign instr_fd = ((pc_fd[30]) ? bios_douta : imem_doutb);

  assign rd1_fd = (rs1_sel == 2'b10) ? wd :
                  (rs1_sel == 2'b01) ? alu_out : rd1; // Forwarding for pipeline register
  assign rd2_fd = (rs2_sel == 2'b10) ? wd :
                  (rs2_sel == 2'b01) ? alu_out : rd2; // Forwarding for pipeline register
  assign rd1_fwd_ex = (rs1_sel_ex) ? wd : rd1_ex; // Forwarding for ALU, Data Gen. and Branch Comp.
  assign rd2_fwd_ex = (rs2_sel_ex) ? wd : rd2_ex; // Forwarding for ALU, Data Gen. and Branch Comp.

  // Configure Arithmetic-Logic Unit inputs           
  assign alu_in_a = (alu_a_sel) ? pc_ex  : (rs1_sel_ex) ? wd : rd1_ex;
  assign alu_in_b = (alu_b_sel) ? imm_ex : (rs2_sel_ex) ? wd : rd2_ex;
  assign funct3   = (instr_ex[6:0] == `OPC_LUI) ? 4'b1000 : (instr_ex[6:0] == `OPC_ARI_RTYPE || instr_ex[6:0] == `OPC_ARI_ITYPE) ? instr_ex[14:12]   : 4'b0; // OP or LOAD/STORE?

  // Configure Data Memory inputs  
  assign dmem_en   = (alu_out == 32'b0) ? 1'b0 : 1'b1;
  assign dmem_addr = alu_out[15:2];
  assign dmem_din  = wr_data;
  assign dmem_we   = (mem_rw) ? wr_mask : 4'b0000;

  // Configure data writeback
  assign wd = (wb_sel == 2'b10) ? pc_wb + 4 :
              (wb_sel == 2'b01) ? alu_wb    : load_extended_data;
  
  always @(*) begin
    if (rst) next_pc = RESET_PC;
    else begin
      case (next_pc_sel)
        2'b00: next_pc = pc_fd + 4;
        2'b01: next_pc = pc_fd + immediate;  // Move to mux with JumpSel
        2'b10: next_pc = rd1_fd + immediate; // Move to mux with JumpSel
        2'b11: next_pc = pc_fd;
        default: next_pc = pc_fd + 4;
      endcase
    end
  end

  always @(posedge clk) begin
    if (rst) begin
      pc_fd <= RESET_PC;
      pc_ex <= 0;
      pc_wb <= 0;

      instr_ex <= 0;
      instr_wb <= 0;

      rd1_ex <= 0;
      rd2_ex <= 0;
      rd2_wb <= 0;
      alu_wb <= 0;

      imm_ex <= 0;
      imm_wb <= 0;

      tohost_csr <= 0;

      prev_instr <= 0;
      instr_count <= 0;
      clk_count <= 0;
    end else begin
      pc_fd <= next_pc;
      pc_ex <= pc_fd;
      pc_wb <= pc_ex;

      instr_ex <= (insert_nop) ? 32'h13 : instr_fd;
      instr_wb <= instr_ex;

      rd1_ex <= rd1_fd;
      rd2_ex <= rd2_fd;
      rd1_wb <= rd1_ex;
      rd2_wb <= rd2_ex;

      alu_wb <= alu_out;

      imm_ex <= immediate;
      imm_wb <= imm_ex;

      if (instr_ex[6:0] == `OPC_CSR) 
        tohost_csr <= (instr_ex[14] == 1) ? imm_ex : rd1_ex;
      
      if(!insert_nop) instr_count <= instr_count + 1;
      clk_count <= clk_count + 1;
    end
  end

  /* Displays debug information */ 
  /*reg [47:0] str_fd;
  reg [47:0] str_ex;
  always @(negedge clk) begin
    case (instr_fd[6:0])
      `OPC_ARI_RTYPE: str_fd = "ARI-R ";
      `OPC_ARI_ITYPE: str_fd = "ARI-I ";
      `OPC_STORE:  str_fd = "STORE ";
      `OPC_LOAD:  str_fd = "LOAD  ";
      `OPC_BRANCH:  str_fd = "BRANCH";
      `OPC_JAL:  str_fd = "JAL   ";
      `OPC_JALR:  str_fd = "JALR  ";
      `OPC_LUI:  str_fd = "LUI   ";
      `OPC_AUIPC:  str_fd = "AUIPC ";
    endcase
    $display("FD: %s %d, %d (%h), %d (%h) | EX: %h | WB: %d (%h) | clk: %d", str_fd, instr_fd[11:7], ra1, rd1_fd, ra2, rd2_fd, instr_ex[11:7], instr_wb[11:7], wd, clk_count);
    $display("[Cycles: %f] ", clk_count);
    $display("[Instructions: %f] ", instr_count);
  end*/
 
endmodule
